#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sstream>

#include "types.h"
#include "fceu.h"

#include "bisqbot.hh"

int BotFrontDisableVideo=0;

#define REFINE_LENGTH  80
#define REPEAT_LENGTH  20
#define MAX_REFINE_PART1 400
#define MAX_REFINE_PART2 500

/* Maximum length of the sequence */
#define ROCKMAN_SEQMAX 65536
static struct BotFrontMode
{
    unsigned char Seq[ROCKMAN_SEQMAX];
    unsigned char BestSeq[ROCKMAN_SEQMAX];
    unsigned long long BestSeqScore;
    int BestSeqLen;
    
    int SeqPos;
    int Cap;
    
    enum { State_Recording, State_Repeating } State;
    int RepeatCount;

} BotFrontMode = { {}, {}, 0, 0, -1, -1, BotFrontMode::State_Recording };

static void DisplaySeq(unsigned char* Seq, unsigned max)
{
    std::stringstream msg;
    for(unsigned a=0; a<max; )
    {
        unsigned c=1, ch=Seq[a];
        while(a+c < max && Seq[a+c] == ch) ++c;
        msg << c; msg << '*';
        
        unsigned b=0;
        if(ch & K_A) { if(b++) msg << ("_"); msg << ("A"); }
        if(ch & K_B) { if(b++) msg << ("_"); msg << ("B"); }
        if(ch & K_SE) { if(b++) msg << ("_"); msg << ("SE"); }
        if(ch & K_ST) { if(b++) msg << ("_"); msg << ("ST"); }
        if(ch & K_U) { if(b++) msg << ("_"); msg << ("U"); }
        if(ch & K_D) { if(b++) msg << ("_"); msg << ("D"); }
        if(ch & K_L) { if(b++) msg << ("_"); msg << ("L"); }
        if(ch & K_R) { if(b++) msg << ("_"); msg << ("R"); }
        if(!b) msg << ("00");
        msg << (", ");
        a += c;
    }
    BotFrontMsg("%s", msg.str().c_str());
}

static void BotFrontGenInput()
{
    BotFrontMode.Seq[BotFrontMode.SeqPos] = BisqBotInput(BotFrontMode.SeqPos);
}

void BotFrontMsg(const char*fmt, ...)
{
    FILE*fp;
    
    va_list ap;
    va_start(ap,fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    
    fp = fopen("rockman.log", "at");
    if(!fp) fp = fopen("rockman.log", "wt");

    va_start(ap,fmt);
    vfprintf(fp, fmt, ap);
    va_end(ap);
    
    fclose(fp);     
}

static void BotFrontReset()
{
    BisqBotReset();
    BotFrontMode.SeqPos = 0;
}

void BotFrontStart()
{
    fprintf(stderr,"BotFrontStart\n");
    FCEUI_SelectState(1, 1);
    FCEUI_SaveState(NULL);
    if(!FCEUI_EmulationPaused()) FCEUI_ToggleEmulationPause();
    BotFrontMode.Cap = ROCKMAN_SEQMAX;
    BotFrontDisableVideo=1;
    BotFrontMsg("--------------- New run  ------------------ \n");
    srand(time(NULL));
    BotFrontReset();
    BotFrontGenInput();
    
    BotFrontMode.BestSeqScore = 0;
    BotFrontMode.BestSeqLen = 0;
    BotFrontMode.State = BotFrontMode::State_Recording;
    BotFrontMode.RepeatCount = 0;
}

void BotFrontEnd()
{
    fprintf(stderr,"BotFrontEnd\n");
    BisqBotReset();
    BotFrontMode.SeqPos = -1;
    BotFrontDisableVideo=0;
} 

void BotFrontToggle()
{
    if(BotFrontMode.SeqPos < 0)
        BotFrontStart();
    else
        BotFrontEnd();
}

unsigned BotFrontGetKeymask()
{
    if(BotFrontMode.SeqPos >= 0)
         return BotFrontMode.Seq[BotFrontMode.SeqPos];
    return 0;
}

/*
    FCEU main loop:
        1. Emulate frame     
        2. Read keyboard input (BotFrontIter)
        3. Update joypad input (BotFrontGetKeymask)

    It is actually possible for keyboard input to be
    read multiple times per frame... Currently we hope
    this don't happen.
*/
/*
    When this function returns, SeqPos is supposed to
    point to _the next emulated frame's_ input.
    
    That is, Seq[SeqPos] tells which frames are being
    held together with the Frame Advance key.
 */
void BotFrontIter()
{
    BisqBotEvaluate(0);
    if(BotFrontMode.SeqPos < 0)
    {
        BisqBotReset();
        return;
    }
    
    /* Upon entry to this function, SeqPos is the frame
     * that was last fed to the emulator.
     */
    switch(BotFrontMode.State)
    {
        case BotFrontMode::State_Repeating:
        {
            unsigned max = REPEAT_LENGTH;
            if(BotFrontMode.BestSeqLen < max) max = BotFrontMode.BestSeqLen;
            
            // Pass out the new frame
            ++BotFrontMode.SeqPos;
            if(BotFrontMode.SeqPos >= max)
            {
                fprintf(stderr, "End repeating!\n");
                // Switch to recording
                BotFrontEnd();
                BotFrontStart();
            }
            /*
              Don't call BotFrontGenInput() here. We're supposed
              to _replay_, not generate new stuff.
            */
            FCEUI_FrameAdvance();
            return;
        }
        case BotFrontMode::State_Recording:
        {
            unsigned frameno = BotFrontMode.SeqPos;
            switch(BisqBotTick(frameno))
            {
                case BisqBotRestart:
                {
                    /*
                        Back to frame 0.
                        - Load state
                        - Reset assumed system state
                        - Provide input
                        - Give permission to emulate next frame.
                    */
                    //BotFrontStateDump();
             Restart:
                    unsigned long long score = BisqBotEvaluate(frameno);
                    if(score > BotFrontMode.BestSeqScore)
                    {
                        fprintf(stderr, "Try %d: score %llu (%u frames)\n", BotFrontMode.RepeatCount,score, frameno);
                        DisplaySeq(BotFrontMode.Seq, BotFrontMode.SeqPos);
                        BotFrontMsg("\n");
                        BotFrontMode.BestSeqScore = score;
                        memcpy(BotFrontMode.BestSeq, BotFrontMode.Seq,
                               sizeof(BotFrontMode.BestSeq));
                        BotFrontMode.BestSeqLen = frameno;
                        /* Do more refining! */
                        BotFrontMode.RepeatCount=0;
                    }
                    else if(score == BotFrontMode.BestSeqScore)
                    {
                        DisplaySeq(BotFrontMode.Seq, BotFrontMode.SeqPos);
                        BotFrontMsg("\n");
                    }
                    
                    if(BotFrontMode.BestSeqScore < 640000000ULL)
                    {
                        // Don't accept this input!
                        --BotFrontMode.RepeatCount;
                    }
                    
                    // Create beginning frame.
                    BotFrontReset();

                    fprintf(stderr, "\r");
                    FCEUI_LoadState(NULL);
                    fprintf(stderr, "[1A\r"); // LoadState will display a line.

                    if(++BotFrontMode.RepeatCount >= MAX_REFINE_PART1)
                    {
                        fprintf(stderr, "Repeating!\n");
                        DisplaySeq(BotFrontMode.BestSeq, REPEAT_LENGTH);
                        BotFrontMsg("\n");
                        BotFrontMode.State = BotFrontMode::State_Repeating;
                        memcpy(BotFrontMode.Seq, BotFrontMode.BestSeq,
                               sizeof(BotFrontMode.Seq));
                    }
                    else
                    {
                        BotFrontGenInput();
                    }
                    FCEUI_FrameAdvance();
                    return;
                }
                case BisqBotContinue:
                {
                    if(frameno+1 >= BotFrontMode.Cap
                    || frameno+1 >= REFINE_LENGTH)
                        goto Restart;

                    /* Create new frame. */
                    ++BotFrontMode.SeqPos;
                    BotFrontGenInput();
                    FCEUI_FrameAdvance();
                    return;
                }
                case BisqBotAccept:
                {
                    /* Report how we got there. */
                    unsigned a;
                    BotFrontMsg("\n***  Bingo in %u frames:\n", BotFrontMode.SeqPos);
                    DisplaySeq(BotFrontMode.Seq, BotFrontMode.SeqPos);
                    BotFrontMsg("\n\n");
                    /*FCEUI_SelectState(3, 1);
                    FCEUI_SaveState(NULL);
                    FCEUI_SelectState(1, 1);*/
                   
                    BotFrontMode.Cap = BotFrontMode.SeqPos+1;
              #if 1
                    /* Stop here. Let the user sort it out. The emulator is paused,
                     * and the movie is ready up to this point.
                     */
                    BotFrontEnd();
                    return;
              #else
                    /* Continue trying to find a faster way to accomplish it. */
                    goto Restart;
              #endif
                }
            }
            break;
        } //recording
    }//switch state
}
